import argparse
import asyncio
from abc import ABC
import signal

from typing import Union, Optional

from ..zmq.base import ZMQRuntime

if False:
    import multiprocessing
    import threading


class AsyncZMQRuntime(ZMQRuntime):
    """
    Runtime procedure in the async manners.

    Base class of :class:`AsyncNewLoopRuntime`.
    """

    def __init__(
        self,
        args: 'argparse.Namespace',
        cancel_event: Optional[
            Union['asyncio.Event', 'multiprocessing.Event', 'threading.Event']
        ] = None,
        **kwargs,
    ):
        super().__init__(args, **kwargs)
        self.is_cancel = cancel_event or asyncio.Event()

    def run_forever(self):
        """Running method to block the main thread."""
        asyncio.run(self._loop_body())

    async def async_cancel(self):
        """An async method to cancel."""
        raise NotImplementedError

    async def async_run_forever(self):
        """The async method to run until it is stopped."""
        raise NotImplementedError

    async def _wait_for_cancel(self):
        """Do NOT override this method when inheriting from :class:`GatewayPea`"""
        # handle terminate signals
        while not self.is_cancel.is_set():
            await asyncio.sleep(0.1)

        await self.async_cancel()

    async def _loop_body(self):
        """Do NOT override this method when inheriting from :class:`GatewayPea`"""
        try:
            await asyncio.gather(self.async_run_forever(), self._wait_for_cancel())
        except asyncio.CancelledError:
            self.logger.warning('received terminate ctrl message from main process')
        await self.async_cancel()


class AsyncNewLoopRuntime(AsyncZMQRuntime, ABC):
    """
    The runtime to start a new event loop.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        try:
            for signame in {'SIGINT', 'SIGTERM'}:
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda *args, **kwargs: self.is_cancel.set(),
                )
        except ValueError as exc:
            self.logger.warning(
                f' The runtime {self.__class__.__name__} will not be able to handle termination signals. '
                f' {repr(exc)}'
            )

        self._loop.run_until_complete(self.async_setup())

    def run_forever(self):
        """
        Running method to block the main thread.

        Run the event loop until a Future is done.
        """
        self._loop.run_until_complete(self._loop_body())

    def teardown(self):
        """Stop and close the event loop."""
        self._loop.stop()
        self._loop.close()
        super().teardown()

    async def async_setup(self):
        """The async method to setup."""
        pass

    # Static methods used by the Pea to communicate with the `Runtime` in the separate process

    @staticmethod
    def cancel(
        cancel_event: Union['multiprocessing.Event', 'threading.Event'], **kwargs
    ):
        """
        Signal the runtime to terminate

        :param cancel_event: the cancel event to set
        :param kwargs: extra keyword arguments
        """
        cancel_event.set()

    @staticmethod
    def activate(**kwargs):
        """
        Activate the runtime, does not apply to these runtimes

        :param kwargs: extra keyword arguments
        """
        # does not apply to this types of runtimes
        pass

    @staticmethod
    def get_control_address(host: str, port: str, **kwargs):
        """
        Get the control address for a runtime with a given host and port

        :param host: the host where the runtime works
        :param port: the control port where the runtime listens
        :param kwargs: extra keyword arguments
        :return: The corresponding control address
        """
        from ...zmq import Zmqlet

        # TODO: I think the control address with ipc from Gateway is not used anymore
        return Zmqlet.get_ctrl_address(host, port, True)[0]

    @staticmethod
    def wait_for_ready_or_shutdown(
        timeout: Optional[float],
        ready_or_shutdown_event: Union['multiprocessing.Event', 'threading.Event'],
        **kwargs,
    ):
        """
        Check if the runtime has successfully started

        :param timeout: The time to wait before readiness or failure is determined
        :param ready_or_shutdown_event: the multiprocessing event to detect if the process failed or succeeded
        :param kwargs: extra keyword arguments

        :return: True if is ready or it needs to be shutdown
        """
        return ready_or_shutdown_event.wait(timeout)



import numpy as np

from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    import scipy

_SPARSE_SCIPY_TYPES = Union[
    'scipy.sparse.csr_matrix',
    'scipy.sparse.csc_matrix',
    'scipy.sparse.bsr_matrix',
    'scipy.sparse.coo_matrix',
]


def pdist(
    x_mat: Union['np.ndarray', _SPARSE_SCIPY_TYPES],
    metric: str,
    is_sparse: bool = False,
) -> 'np.ndarray':
    """Computes Pairwise distances between observations in n-dimensional space.

    :param x_mat: Union['np.ndarray','scipy.sparse.csr_matrix', 'scipy.sparse.coo_matrix'] of ndim 2
    :param metric: string describing the metric type
    :param is_sparse: boolean stating if computation must be done with sparse data
    :return: np.ndarray of ndim 2
    """
    return cdist(x_mat, x_mat, metric, is_sparse)


def cdist(
    x_mat: Union['np.ndarray', _SPARSE_SCIPY_TYPES],
    y_mat: Union['np.ndarray', _SPARSE_SCIPY_TYPES],
    metric: str,
    is_sparse: bool = False,
) -> 'np.ndarray':

    """Computes the pairwise distance between each row of X and each row on Y according to `metric`.
    - Let `n_x = x_mat.shape[0]`
    - Let `n_y = y_mat.shape[0]`
    - Returns a matrix `dist` of shape `(n_x, n_y)` with `dist[i,j] = metric(x_mat[i], y_mat[j])`.
    :param x_mat: numpy or scipy array of ndim 2
    :param y_mat: numpy or scipy array of ndim 2
    :param metric: string describing the metric type
    :param is_sparse: boolean describing if data type is sparse
    :return: np.ndarray of ndim 2
    """
    if metric == 'cosine':
        if is_sparse:
            dists = sparse_cosine(x_mat, y_mat)
        else:
            dists = cosine(x_mat, y_mat)

    elif metric == 'sqeuclidean':
        if is_sparse:
            dists = sparse_sqeuclidean(x_mat, y_mat)
        else:
            dists = sqeuclidean(x_mat, y_mat)

    elif metric == 'euclidean':
        if is_sparse:
            dists = np.sqrt(sparse_sqeuclidean(x_mat, y_mat))
        else:
            dists = np.sqrt(sqeuclidean(x_mat, y_mat))
    else:
        raise ValueError(f'Input metric={metric} not valid')
    return dists


def cosine(x_mat: 'np.ndarray', y_mat: 'np.ndarray') -> 'np.ndarray':
    """Cosine distance between each row in x_mat and each row in y_mat.
    :param x_mat: np.ndarray with ndim=2
    :param y_mat: np.ndarray with ndim=2
    :return: np.ndarray  with ndim=2
    """
    return 1 - np.dot(x_mat, y_mat.T) / np.outer(
        np.linalg.norm(x_mat, axis=1), np.linalg.norm(y_mat, axis=1)
    )


def sqeuclidean(x_mat: 'np.ndarray', y_mat: 'np.ndarray') -> 'np.ndarray':
    """squared Euclidean distance between each row in x_mat and each row in y_mat.
    :param x_mat: np.ndarray with ndim=2
    :param y_mat: np.ndarray with ndim=2
    :return: np.ndarray with ndim=2
    """
    return (
        np.sum(y_mat ** 2, axis=1)
        + np.sum(x_mat ** 2, axis=1)[:, np.newaxis]
        - 2 * np.dot(x_mat, y_mat.T)
    )


def sparse_cosine(
    x_mat: _SPARSE_SCIPY_TYPES, y_mat: _SPARSE_SCIPY_TYPES
) -> 'np.ndarray':
    """Cosine distance between each row in x_mat and each row in y_mat.
    :param x_mat:  scipy.sparse like array with ndim=2
    :param y_mat:  scipy.sparse like array with ndim=2
    :return: np.ndarray  with ndim=2
    """
    from scipy.sparse.linalg import norm

    # we need the np.asarray otherwise we get a np.matrix object that iterates differently
    return np.asarray(
        1 - x_mat.dot(y_mat.T) / np.outer(norm(x_mat, axis=1), norm(y_mat, axis=1))
    )


def sparse_sqeuclidean(
    x_mat: _SPARSE_SCIPY_TYPES, y_mat: _SPARSE_SCIPY_TYPES
) -> 'np.ndarray':
    """Cosine distance between each row in x_mat and each row in y_mat.
    :param x_mat:  scipy.sparse like array with ndim=2
    :param y_mat:  scipy.sparse like array with ndim=2
    :return: np.ndarray  with ndim=2
    """
    # we need the np.asarray otherwise we get a np.matrix object that iterates differently
    return np.asarray(
        y_mat.power(2).sum(axis=1).flatten()
        + x_mat.power(2).sum(axis=1)
        - 2 * x_mat.dot(y_mat.T)
    )



from typing import Dict

from pydantic import BaseModel, Field

from .base import StoreItem, StoreStatus
from .id import DaemonID


class ContainerArguments(BaseModel):
    """Pydantic model for ContainerArguments"""

    object: Dict
    command: str


class ContainerMetadata(BaseModel):
    """Pydantic model for ContainerMetadata"""

    container_id: str
    container_name: str
    image_id: str
    network: str
    ports: Dict
    uri: str


class ContainerItem(StoreItem):
    """Pydantic model for ContainerItem"""

    metadata: ContainerMetadata
    arguments: ContainerArguments
    workspace_id: DaemonID


class ContainerStoreStatus(StoreStatus):
    """Pydantic model for ContainerStoreStatus"""

    items: Dict[DaemonID, ContainerItem] = Field(default_factory=dict)



"""Module for wrapping Jina Hub API calls."""

import argparse
import hashlib
import json
import os
import random
from pathlib import Path
from typing import Optional, Dict
from urllib.parse import urlencode

from . import HubExecutor
from .helper import (
    archive_package,
    download_with_resume,
    parse_hub_uri,
    get_hubble_url,
    upload_file,
    disk_cache_offline,
)
from .helper import install_requirements
from .hubapi import install_local, resolve_local, load_secret, dump_secret, get_lockfile
from .. import __resources_path__
from ..helper import get_full_version, ArgNamespace
from ..importer import ImportExtensions
from ..logging.logger import JinaLogger
from ..parsers.hubble import set_hub_parser

_cache_file = Path.home().joinpath('.jina', 'disk_cache.db')


class HubIO:
    """:class:`HubIO` provides the way to interact with Jina Hub registry.
    You can use it with CLI to package a directory into a Jina Hub and publish it to the world.
    Examples:
        - :command:`jina hub push my_executor/` to push the executor package to Jina Hub
        - :command:`jina hub pull UUID8` to download the executor identified by UUID8

    To create a :class:`HubIO` object, simply:

        .. highlight:: python
        .. code-block:: python
            hubio = HubIO(args)

    :param args: arguments
    """

    def __init__(self, args: Optional[argparse.Namespace] = None, **kwargs):
        if args and isinstance(args, argparse.Namespace):
            self.args = args
        else:
            self.args = ArgNamespace.kwargs2namespace(kwargs, set_hub_parser())
        self.logger = JinaLogger(self.__class__.__name__, **vars(args))

        with ImportExtensions(required=True):
            import rich
            import cryptography
            import filelock

            assert rich  #: prevent pycharm auto remove the above line
            assert cryptography
            assert filelock

    def _load_docker_client(self):
        with ImportExtensions(required=True):
            import docker
            from docker import APIClient

            # low-level client
            self._raw_client = APIClient(base_url='unix://var/run/docker.sock')

    @staticmethod
    def _get_request_header() -> Dict:
        """Return the header of request.

        :return: request header
        """
        metas, envs = get_full_version()

        header = {
            **{f'jinameta-{k}': str(v) for k, v in metas.items()},
            **envs,
        }
        return header

    def new(self) -> None:
        """Create a new executor folder interactively."""

        from rich import print, box
        from rich.prompt import Prompt, Confirm
        from rich.panel import Panel
        from rich.table import Table
        from rich.console import Console
        from rich.progress import track
        from rich.syntax import Syntax

        console = Console()

        print(
            Panel.fit(
                '''
[bold green]Executor[/bold green] is how Jina processes [bold]Document[/bold]. 

This guide helps you to create your own Executor in 30 seconds.''',
                title='Create New Executor',
            )
        )

        exec_name = Prompt.ask(
            ':grey_question: What is the [bold]name[/bold] of your executor?\n'
            '[dim]CamelCase is required[/dim]',
            default=f'MyExecutor{random.randint(0, 100)}',
        )

        exec_path = Prompt.ask(
            ':grey_question: [bold]Which folder[/bold] to store your executor?',
            default=os.path.join(os.getcwd(), exec_name),
        )

        exec_description = '{{}}'
        exec_keywords = '{{}}'
        exec_author = '{{}}'
        exec_url = '{{}}'

        is_dockerfile = False

        if Confirm.ask(
            '[green]That\'s all we need to create an Executor![/green]\n'
            ':grey_question: Or do you want to proceed to advanced configuration',
            default=False,
        ):
            exec_description = (
                Prompt.ask(
                    ':grey_question: Please give a [bold]short description[/bold] of your executor?\n'
                    f'[dim]Example: {exec_name} embeds images into 128-dim vectors using ResNet.[/dim]'
                )
                or exec_description
            )

            exec_author = (
                Prompt.ask(
                    ':grey_question: What is the [bold]author name[/bold]?\n'
                    f'[dim]Example: John Doe[/dim]'
                )
                or exec_author
            )

            exec_keywords = (
                Prompt.ask(
                    ':grey_question: Please give some [bold]keywords[/bold] to help people search your executor [dim](separated by space)[/dim]\n'
                    f'[dim]Example: image cv embedding encoding resnet[/dim]'
                )
                or exec_keywords
            )

            exec_url = (
                Prompt.ask(
                    ':grey_question: What is the [bold]URL[/bold] for documentation/reference/more info?\n'
                    f'[dim]Example: https://docs.jina.ai[/dim]'
                )
                or exec_url
            )

            print(
                Panel.fit(
                    '''
[bold]Dockerfile[/bold] describes how this executor will be built. It is useful when 
your executor has non-trivial dependencies or must be run under certain environment. 

- If the [bold]Dockerfile[/bold] is missing, Jina automatically generates one for you. 
- If you provide one, then Jina will respect the given [bold]Dockerfile[/bold].''',
                    title='[Optional] [bold]Dockerfile[/bold]',
                    width=80,
                )
            )

            is_dockerfile = Confirm.ask(
                ':grey_question: Do you need to write your own [bold]Dockerfile[/bold] instead of the auto-generated one?',
                default=False,
            )

            print('[green]That\'s all we need to create an Executor![/green]')

        def mustache_repl(srcs):
            for src in track(
                srcs, description=f'Creating {exec_name}...', total=len(srcs)
            ):
                with open(
                    os.path.join(__resources_path__, 'executor-template', src)
                ) as fp, open(os.path.join(exec_path, src), 'w') as fpw:
                    f = (
                        fp.read()
                        .replace('{{exec_name}}', exec_name)
                        .replace('{{exec_description}}', exec_description)
                        .replace('{{exec_keywords}}', exec_keywords)
                        .replace('{{exec_author}}', exec_author)
                        .replace('{{exec_url}}', exec_url)
                    )

                    f = [
                        v + '\n' for v in f.split('\n') if not ('{{' in v or '}}' in v)
                    ]
                    fpw.writelines(f)

        Path(exec_path).mkdir(parents=True, exist_ok=True)
        pkg_files = [
            'executor.py',
            'manifest.yml',
            'README.md',
            'requirements.txt',
            'config.yml',
        ]

        if is_dockerfile:
            pkg_files.append('Dockerfile')

        mustache_repl(pkg_files)

        table = Table(box=box.SIMPLE)
        table.add_column('Filename', style='cyan', no_wrap=True)
        table.add_column('Description', no_wrap=True)

        table.add_row('executor.py', 'The main logic file of the Executor.')
        table.add_row(
            'config.yml',
            'The YAML config file of the Executor. You can define [bold]__init__[/bold] arguments using [bold]with[/bold] keyword.',
        )
        table.add_row(
            '',
            Panel(
                Syntax(
                    f'''
jtype: {exec_name}
with:
    foo: 1
    bar: hello
py_modules:
    - executor.py
                ''',
                    'yaml',
                    theme='monokai',
                    line_numbers=True,
                    word_wrap=True,
                ),
                title='config.yml',
                width=50,
                expand=False,
            ),
        )
        table.add_row('README.md', 'The usage of the Executor.')
        table.add_row('requirements.txt', 'The Python dependencies of the Executor.')
        table.add_row(
            'manifest.yml',
            'The annotations of the Executor for getting better appealing on Jina Hub.',
        )

        field_table = Table(box=box.SIMPLE)
        field_table.add_column('Field', style='cyan', no_wrap=True)
        field_table.add_column('Description', no_wrap=True)
        field_table.add_row('name', 'Human-readable title of the Executor')
        field_table.add_row('alias', 'The unique identifier in Jina Hub')
        field_table.add_row('description', 'Human-readable description of the Executor')
        field_table.add_row('author', 'The author of the Executor')
        field_table.add_row('url', 'URL to find more information on the Executor')
        field_table.add_row('keywords', 'Keywords that help user find the Executor')

        table.add_row('', field_table)

        if is_dockerfile:
            table.add_row(
                'Dockerfile',
                'The Dockerfile describes how this executor will be built.',
            )

        final_table = Table(box=None)

        final_table.add_row(
            'Congrats! You have successfully created an Executor! Here are the next steps:'
        )

        p0 = Panel(
            Syntax(
                f'cd {exec_path}\nls',
                'console',
                theme='monokai',
                line_numbers=True,
                word_wrap=True,
            ),
            title='1. Checkout the Generated Executor',
            width=120,
            expand=False,
        )

        p1 = Panel(
            table,
            title='2. Understand Folder Structure',
            width=120,
            expand=False,
        )

        p2 = Panel(
            Syntax(
                f'jina hub push {exec_path}',
                'console',
                theme='monokai',
                line_numbers=True,
                word_wrap=True,
            ),
            title='3. Share it to Jina Hub',
            width=120,
            expand=False,
        )

        final_table.add_row(p0)
        final_table.add_row(p1)
        final_table.add_row(p2)

        p = Panel(
            final_table,
            title=':tada: Next Steps',
            width=130,
            expand=False,
        )
        console.print(p)

    def push(self) -> None:
        """Push the executor pacakge to Jina Hub."""

        from rich.console import Console

        work_path = Path(self.args.path)

        exec_tags = None
        if self.args.tag:
            exec_tags = ','.join(self.args.tag)

        dockerfile = None
        if self.args.docker_file:
            dockerfile = Path(self.args.docker_file)
            if not dockerfile.exists():
                raise Exception(f'The given Dockerfile `{dockerfile}` does not exist!')
            if dockerfile.parent != work_path:
                raise Exception(
                    f'The Dockerfile must be placed at the given folder `{work_path}`'
                )

            dockerfile = dockerfile.relative_to(work_path)

        console = Console()
        with console.status(f'Pushing `{self.args.path}` ...') as st:
            req_header = self._get_request_header()
            try:
                st.update(f'Packaging {self.args.path} ...')
                md5_hash = hashlib.md5()
                bytesio = archive_package(work_path)
                content = bytesio.getvalue()
                md5_hash.update(content)
                md5_digest = md5_hash.hexdigest()

                # upload the archived package
                form_data = {
                    'public': 'True' if getattr(self.args, 'public', None) else 'False',
                    'private': 'True'
                    if getattr(self.args, 'private', None)
                    else 'False',
                    'md5sum': md5_digest,
                }

                if exec_tags:
                    form_data['tags'] = exec_tags

                if dockerfile:
                    form_data['dockerfile'] = str(dockerfile)

                uuid8, secret = load_secret(work_path)
                if self.args.force or uuid8:
                    form_data['force'] = self.args.force or uuid8
                if self.args.secret or secret:
                    form_data['secret'] = self.args.secret or secret

                method = 'put' if ('force' in form_data) else 'post'

                st.update(f'Connecting Hubble ...')
                hubble_url = get_hubble_url()

                # upload the archived executor to Jina Hub
                st.update(f'Uploading ...')
                resp = upload_file(
                    hubble_url,
                    'filename',
                    content,
                    dict_data=form_data,
                    headers=req_header,
                    stream=True,
                    method=method,
                )

                result = None
                for stream_line in resp.iter_lines():
                    stream_msg = json.loads(stream_line)

                    if 'stream' in stream_msg:
                        console.print(f'=> {stream_msg["stream"]}')
                    elif 'status' in stream_msg:
                        st.update(f'{stream_msg["status"]}')
                    elif 'result' in stream_msg:
                        result = stream_msg['result']
                        break

                if result is None:
                    raise Exception('Unknown Error')
                elif not result.get('data', None):
                    raise Exception(result.get('message', 'Unknown Error'))
                elif 200 <= result['statusCode'] < 300:
                    new_uuid8, new_secret = self._prettyprint_result(console, result)
                    if new_uuid8 != uuid8 or new_secret != secret:
                        dump_secret(work_path, new_uuid8, new_secret)
                elif result['message']:
                    raise Exception(result['message'])
                elif resp.text:
                    # NOTE: sometimes resp.text returns empty
                    raise Exception(resp.text)
                else:
                    resp.raise_for_status()

            except Exception as e:  # IO related errors
                self.logger.error(
                    f'Error while pushing session_id={req_header["jinameta-session-id"]}: '
                    f'\n{e!r}'
                )
                raise e

    def _prettyprint_result(self, console, result):
        # TODO: only support single executor now

        from rich.table import Table
        from rich import box

        data = result.get('data', None)
        image = data['executors'][0]
        uuid8 = image['id']
        secret = image['secret']
        visibility = image['visibility']

        table = Table(box=box.SIMPLE)
        table.add_column('Key', no_wrap=True)
        table.add_column('Value', style='cyan', no_wrap=True)
        table.add_row(':key: ID', uuid8)
        if 'alias' in image:
            table.add_row(':name_badge: Alias', image['alias'])
        table.add_row(':lock: Secret', secret)
        table.add_row(
            '',
            ':point_up:️ [bold red]Please keep this token in a safe place!',
        )
        table.add_row(':eyes: Visibility', visibility)
        table.add_row(':whale: DockerHub', f'https://hub.docker.com/r/jinahub/{uuid8}/')
        console.print(table)

        presented_id = image.get('alias', uuid8)
        usage = (
            f'{presented_id}' if visibility == 'public' else f'{presented_id}:{secret}'
        )

        if not self.args.no_usage:
            self._get_prettyprint_usage(console, usage)

        return uuid8, secret

    def _get_prettyprint_usage(self, console, usage):
        from rich.panel import Panel
        from rich.syntax import Syntax

        flow_plain = f'''from jina import Flow

f = Flow().add(uses='jinahub://{usage}')

with f:
    ...'''

        flow_docker = f'''from jina import Flow

f = Flow().add(uses='jinahub+docker://{usage}')

with f:
    ...'''

        cli_plain = f'jina executor --uses jinahub://{usage}'

        cli_docker = f'jina executor --uses jinahub+docker://{usage}'

        p1 = Panel(
            Syntax(
                flow_plain, 'python', theme='monokai', line_numbers=True, word_wrap=True
            ),
            title='Flow usage',
            width=80,
            expand=False,
        )
        p2 = Panel(
            Syntax(
                flow_docker,
                'python',
                theme='monokai',
                line_numbers=True,
                word_wrap=True,
            ),
            title='Flow usage via Docker',
            width=80,
            expand=False,
        )

        p3 = Panel(
            Syntax(
                cli_plain, 'console', theme='monokai', line_numbers=True, word_wrap=True
            ),
            title='CLI usage',
            width=80,
            expand=False,
        )
        p4 = Panel(
            Syntax(
                cli_docker,
                'console',
                theme='monokai',
                line_numbers=True,
                word_wrap=True,
            ),
            title='CLI usage via Docker',
            width=80,
            expand=False,
        )

        console.print(p1, p2, p3, p4)

    @staticmethod
    @disk_cache_offline(cache_file=str(_cache_file))
    def _fetch_meta(
        name: str,
        tag: Optional[str] = None,
        secret: Optional[str] = None,
    ) -> HubExecutor:
        """Fetch the executor meta info from Jina Hub.
        :param name: the UUID/Alias of the executor
        :param tag: the version tag of the executor
        :param secret: the access secret of the executor
        :return: meta of executor
        """

        with ImportExtensions(required=True):
            import requests

        pull_url = get_hubble_url() + f'/{name}/?'
        path_params = {}
        if secret:
            path_params['secret'] = secret
        if tag:
            path_params['tag'] = tag
        if path_params:
            pull_url += urlencode(path_params)

        resp = requests.get(pull_url, headers=HubIO._get_request_header())
        if resp.status_code != 200:
            if resp.text:
                raise Exception(resp.text)
            resp.raise_for_status()

        resp = resp.json()

        return HubExecutor(
            uuid=resp['id'],
            alias=resp.get('alias', None),
            sn=resp.get('sn', None),
            tag=resp['tag'],
            visibility=resp['visibility'],
            image_name=resp['image'],
            archive_url=resp['package']['download'],
            md5sum=resp['package']['md5'],
        )

    def pull(self) -> str:
        """Pull the executor package from Jina Hub.

        :return: the `uses` string
        """
        from rich.console import Console
        from rich.progress import Progress, DownloadColumn, BarColumn

        console = Console()
        cached_zip_file = None
        usage = None

        try:
            with console.status(f'Pulling {self.args.uri}...') as st:
                scheme, name, tag, secret = parse_hub_uri(self.args.uri)

                st.update(f'Fetching meta data of {name}...')
                executor = HubIO._fetch_meta(name, tag=tag, secret=secret)
                usage = (
                    f'{executor.uuid}'
                    if executor.visibility == 'public'
                    else f'{executor.uuid}:{secret}'
                )

            if scheme == 'jinahub+docker':
                self._load_docker_client()

                with Progress(
                    "[progress.description]{task.description}",
                    BarColumn(),
                    DownloadColumn(),
                    "[progress.percentage]{task.percentage:>3.0f}%",
                    console=console,
                    transient=True,
                ) as progress:

                    try:
                        tasks = {}
                        for log in self._raw_client.pull(
                            executor.image_name, stream=True, decode=True
                        ):
                            if 'status' not in log:
                                continue
                            status = log['status']
                            status_id = log.get('id', None)
                            pg_detail = log.get('progressDetail', None)

                            if (pg_detail is None) or (status_id is None):
                                console.print(status)
                                continue

                            task = tasks.get(
                                status_id, progress.add_task(status, total=0)
                            )
                            current = pg_detail['current']
                            total = pg_detail['total']
                            progress.update(
                                task,
                                completed=current,
                                total=total,
                                description=status,
                            )

                            if total and current == total:
                                progress.stop_task(task)

                    except Exception as ex:
                        raise ex

                return f'docker://{executor.image_name}'
            elif scheme == 'jinahub':
                import filelock

                with filelock.FileLock(get_lockfile(), timeout=-1):
                    try:
                        pkg_path, pkg_dist_path = resolve_local(executor)
                        # check serial number to upgrade
                        sn_file_path = pkg_dist_path / f'PKG-SN-{executor.sn or 0}'
                        if (not sn_file_path.exists()) and any(
                            pkg_dist_path.glob('PKG-SN-*')
                        ):
                            raise FileNotFoundError(f'{pkg_path} need to be upgraded')
                        if self.args.install_requirements:
                            requirements_file = pkg_dist_path / 'requirements.txt'
                            if requirements_file.exists():
                                install_requirements(requirements_file)
                        return f'{pkg_path / "config.yml"}'
                    except FileNotFoundError:
                        pass  # have not been downloaded yet, download for the first time
                    # download the package
                    cache_dir = Path(
                        os.environ.get(
                            'JINA_HUB_CACHE_DIR',
                            Path.home().joinpath('.cache', 'jina'),
                        )
                    )
                    cache_dir.mkdir(parents=True, exist_ok=True)

                    with console.status(f'Downloading {name} ...') as st:
                        cached_zip_file = download_with_resume(
                            executor.archive_url,
                            cache_dir,
                            f'{executor.uuid}-{executor.md5sum}.zip',
                            md5sum=executor.md5sum,
                        )

                        st.update(f'Unpacking {name} ...')
                        install_local(
                            cached_zip_file,
                            executor,
                            install_deps=self.args.install_requirements,
                        )

                        pkg_path, _ = resolve_local(executor)
                    return f'{pkg_path / "config.yml"}'
            else:
                raise ValueError(f'{self.args.uri} is not a valid scheme')
        except Exception as e:
            self.logger.error(f'Error while pulling {self.args.uri}: \n{e!r}')
            raise e
        finally:
            # delete downloaded zip package if existed
            if cached_zip_file is not None:
                cached_zip_file.unlink()

            if not self.args.no_usage and usage:
                self._get_prettyprint_usage(console, usage)



import os
import json
import asyncio
from http import HTTPStatus
from typing import Dict, Optional, Union

import aiohttp

from jina import __resources_path__
from jina.helper import run_async
from jina.logging.logger import JinaLogger

from .mixin import AsyncToSyncMixin
from ..models.id import DaemonID, daemonize
from ..helper import error_msg_from, if_alive


class AsyncBaseClient:
    """
    JinaD baseclient (Async)

    :param uri: the uri of ``jinad`` instance
    :param logger: jinad logger
    :param timeout: stop waiting for a response after a given number of seconds with the timeout parameter.
    """

    _kind = ''
    _endpoint = '/'

    def __init__(
        self,
        uri: str,
        logger: JinaLogger,
        timeout: int = None,
    ):
        self._logger = logger
        self.timeout = timeout
        self.http_uri = f'http://{uri}'
        self.store_api = f'{self.http_uri}{self._endpoint}'
        self.logstream_api = f'ws://{uri}/logstream'

    @if_alive
    async def alive(self) -> bool:
        """Check if JinaD is alive

        :return: True if JinaD is alive at remote else false
        """
        async with aiohttp.request(method='GET', url=self.http_uri) as response:
            return response.status == HTTPStatus.OK

    @if_alive
    async def status(self) -> Optional[Dict]:
        """Get status of remote JinaD

        :return: dict status of remote JinaD
        """
        async with aiohttp.request(
            method='GET', url=f'{self.http_uri}/status'
        ) as response:
            if response.status == HTTPStatus.OK:
                return await response.json()

    @if_alive
    async def get(self, id: Union[str, DaemonID]) -> Optional[Union[str, Dict]]:
        """Get status of the remote object

        :param id: identity of the Pea/Pod
        :return: response if the remote Flow/Pea/Pod/Workspace exists
        """

        async with aiohttp.request(
            method='GET',
            url=f'{self.store_api}/{daemonize(id, self._kind)}',
        ) as response:
            response_json = await response.json()
            if response.status == HTTPStatus.UNPROCESSABLE_ENTITY:
                self._logger.error(
                    f'validation error in the request: {error_msg_from(response_json)}'
                )
                return response_json['body']
            elif response.status == HTTPStatus.NOT_FOUND:
                self._logger.error(
                    f'couldn\'t find {id} in remote {self._kind.title()} store'
                )
                return response_json['detail']
            else:
                return response_json

    @if_alive
    async def list(self) -> Dict:
        """List all objects in the store

        :return: json response of the remote Flow/Pea/Pod/Workspace status
        """
        async with aiohttp.request(method='GET', url=self.store_api) as response:
            if response.status == HTTPStatus.OK:
                response_json = await response.json()
                self._logger.success(
                    f'found {len(response_json.get("items", []))} {self._kind.title()}(s) in store'
                )
                return (
                    response_json['items']
                    if 'items' in response_json
                    else response_json
                )
            else:
                self._logger.error(
                    f'got response {response.status} while listing all {self._kind}s'
                )

    async def create(self, *args, **kwargs) -> Dict:
        """Create a Workspace/Flow/Pea/Pod on remote.
        Must be implemented by the inherited class.

        # noqa: DAR101
        # noqa: DAR102
        """
        raise NotImplementedError

    async def update(self, *args, **kwargs) -> Dict:
        """Update a Workspace/Flow/Pea/Pod on remote.
        Must be implemented by the inherited class.

        # noqa: DAR101
        # noqa: DAR102
        """
        raise NotImplementedError

    async def delete(self, id: DaemonID, *args, **kwargs) -> str:
        """Delete a Workspace/Flow/Pea/Pod on remote.
        Must be implemented by the inherited class.

        # noqa: DAR101
        # noqa: DAR102
        """
        raise NotImplementedError

    async def logstream(self, id: 'DaemonID', timeout: float = None):
        """Websocket log stream from remote Workspace/Flow/Pea/Pod

        :param id: identity of the Workspace/Flow/Pea/Pod
        :param timeout: timeout in seconds for the logstream
        """
        remote_log_config = os.path.join(__resources_path__, 'logging.remote.yml')
        remote_loggers: Dict[str, 'JinaLogger'] = {}
        try:
            id = daemonize(id)
        except TypeError:
            self._logger.error(f'invalid id {id} passed for logstreaming, exiting..')
            return
        url = f'{self.logstream_api}/{id}'
        try:
            async with aiohttp.ClientSession() as session:
                while True:
                    try:
                        async with session.ws_connect(url) as websocket:
                            async for log_line in websocket:
                                try:
                                    if not log_line:
                                        continue
                                    json_log_line = log_line.json()
                                    name = json_log_line.get('name')
                                    if name not in remote_loggers:
                                        remote_loggers[name] = JinaLogger(
                                            context=json_log_line.get('host'),
                                            log_config=remote_log_config,
                                        )

                                    remote_loggers[name].debug(
                                        '{host} {name} {type} {message}'.format_map(
                                            json_log_line
                                        )
                                    )
                                except json.decoder.JSONDecodeError:
                                    continue
                    except aiohttp.WSServerHandshakeError as e:
                        self._logger.error(
                            f'log streaming failed, you won\'t see logs on the remote\n Reason: {e!r}'
                        )
                        continue
                    except asyncio.CancelledError:
                        self._logger.debug(
                            f'successfully cancelled log streaming task for {id}'
                        )
                        break
        finally:
            for logger in remote_loggers.values():
                logger.close()


class BaseClient(AsyncToSyncMixin, AsyncBaseClient):
    """JinaD baseclient"""



from http import HTTPStatus
from typing import Dict, Optional, TYPE_CHECKING, Tuple, Union

import aiohttp

from ..models.id import daemonize
from .base import AsyncBaseClient
from .mixin import AsyncToSyncMixin
from ..helper import error_msg_from, if_alive

if TYPE_CHECKING:
    from ..models import DaemonID


class AsyncPeaClient(AsyncBaseClient):
    """Async Client to create/update/delete Peas on remote JinaD"""

    _kind = 'pea'
    _endpoint = '/peas'

    @if_alive
    async def arguments(self) -> Optional[Dict]:
        """Get all arguments accepted by a remote Pea/Pod

        :return: dict arguments of remote JinaD
        """
        async with aiohttp.request(
            method='GET', url=f'{self.store_api}/arguments'
        ) as response:
            if response.status == HTTPStatus.OK:
                return await response.json()

    @if_alive
    async def create(
        self, workspace_id: Union[str, 'DaemonID'], payload: Dict
    ) -> Tuple[bool, str]:
        """Create a remote Pea / Pod

        :param workspace_id: id of workspace where the Pea would live in
        :param payload: json payload
        :return: (True if pea creation succeeded) and (the identity of the spawned Pea/Pod or, error message)
        """

        async with aiohttp.request(
            method='POST',
            url=self.store_api,
            params={'workspace_id': daemonize(workspace_id)},
            json=payload,
        ) as response:
            response_json = await response.json()
            if response.status == HTTPStatus.CREATED:
                self._logger.success(
                    f'successfully created a {self._kind.title()} {response_json} in workspace {workspace_id}'
                )
                return True, response_json
            elif response.status == HTTPStatus.UNPROCESSABLE_ENTITY:
                error_msg = f'validation error in the payload: {response_json["detail"][0]["msg"]}'
                self._logger.error(error_msg)
                return False, error_msg
            else:
                error_msg = error_msg_from(response_json)
                self._logger.error(
                    f'{self._kind.title()} creation failed as: {error_msg}'
                )
                return False, error_msg

    @if_alive
    async def delete(self, id: Union[str, 'DaemonID'], **kwargs) -> bool:
        """Delete a remote Pea/Pod

        :param id: the identity of the Pea/Pod
        :param kwargs: keyword arguments
        :return: True if the deletion is successful
        """

        async with aiohttp.request(
            method='DELETE', url=f'{self.store_api}/{daemonize(id, self._kind)}'
        ) as response:
            response_json = await response.json()
            if response.status != HTTPStatus.OK:
                self._logger.error(
                    f'deletion of {self._kind.title()} {id} failed: {error_msg_from(response_json)}'
                )
            return response.status == HTTPStatus.OK


class PeaClient(AsyncToSyncMixin, AsyncPeaClient):
    """Client to create/update/delete Peas on remote JinaD"""



import gzip
import os
import random
import urllib.request
import webbrowser
from collections import defaultdict

import numpy as np

from jina import Document
from jina.helper import colored
from jina.logging.predefined import default_logger
from jina.logging.profile import ProgressBar

result_html = []
top_k = 0
num_docs_evaluated = 0
evaluation_value = defaultdict(float)


def _get_groundtruths(target, pseudo_match=True):
    # group doc_ids by their labels
    a = np.squeeze(target['index-labels']['data'])
    a = np.stack([a, np.arange(len(a))], axis=1)
    a = a[a[:, 0].argsort()]
    lbl_group = np.split(a[:, 1], np.unique(a[:, 0], return_index=True)[1][1:])

    # each label has one groundtruth, i.e. all docs that have the same label are considered as matches
    groundtruths = {lbl: Document() for lbl in range(10)}
    for lbl, doc_ids in enumerate(lbl_group):
        if not pseudo_match:
            # full-match, each doc has 6K matches
            for doc_id in doc_ids:
                match = Document()
                match.tags['id'] = int(doc_id)
                groundtruths[lbl].matches.append(match)
        else:
            # pseudo-match, each doc has only one match, but this match's id is a list of 6k elements
            match = Document()
            match.tags['id'] = doc_ids.tolist()
            groundtruths[lbl].matches.append(match)

    return groundtruths


def index_generator(num_docs: int, target: dict):
    """
    Generate the index data.

    :param num_docs: Number of documents to be indexed.
    :param target: Dictionary which stores the data paths
    :yields: index data
    """
    for internal_doc_id in range(num_docs):
        # x_blackwhite.shape is (28,28)
        x_blackwhite = 255 - target['index']['data'][internal_doc_id]
        # x_color.shape is (28,28,3)
        x_color = np.stack((x_blackwhite,) * 3, axis=-1)
        d = Document(content=x_color)
        d.tags['id'] = internal_doc_id
        yield d


def query_generator(num_docs: int, target: dict, with_groundtruth: bool = True):
    """
    Generate the query data.

    :param num_docs: Number of documents to be queried
    :param target: Dictionary which stores the data paths
    :param with_groundtruth: True if want to include labels into query data
    :yields: query data
    """
    gts = _get_groundtruths(target)
    for _ in range(num_docs):
        num_data = len(target['query-labels']['data'])
        idx = random.randint(0, num_data - 1)
        # x_blackwhite.shape is (28,28)
        x_blackwhite = 255 - target['query']['data'][idx]
        # x_color.shape is (28,28,3)
        x_color = np.stack((x_blackwhite,) * 3, axis=-1)
        d = Document(content=x_color)

        if with_groundtruth:
            gt = gts[target['query-labels']['data'][idx][0]]
            yield d, gt
        else:
            yield d


def print_result(resp):
    """
    Callback function to receive results.

    :param resp: returned response with data
    """
    global top_k
    global evaluation_value
    for d in resp.docs:
        vi = d.uri
        result_html.append(f'<tr><td><img src="{vi}"/></td><td>')
        top_k = len(d.matches)
        for kk in d.matches:
            kmi = kk.uri
            result_html.append(
                f'<img src="{kmi}" style="opacity:{kk.scores["cosine"].value}"/>'
            )
        result_html.append('</td></tr>\n')

        # update evaluation values
        # as evaluator set to return running avg, here we can simply replace the value
        for k, evaluation in d.evaluations.items():
            evaluation_value[k] = evaluation.value


def write_html(html_path):
    """
    Method to present results in browser.

    :param html_path: path of the written html
    """

    with open(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), 'demo.html')
    ) as fp, open(html_path, 'w') as fw:
        t = fp.read()
        t = t.replace('{% RESULT %}', '\n'.join(result_html))
        t = t.replace(
            '{% PRECISION_EVALUATION %}',
            '{:.2f}%'.format(evaluation_value['Precision'] * 100.0),
        )
        t = t.replace(
            '{% RECALL_EVALUATION %}',
            '{:.2f}%'.format(evaluation_value['Recall'] * 100.0),
        )
        t = t.replace('{% TOP_K %}', str(top_k))
        fw.write(t)

    url_html_path = 'file://' + os.path.abspath(html_path)

    try:
        webbrowser.open(url_html_path, new=2)
    except:
        pass  # intentional pass, browser support isn't cross-platform
    finally:
        default_logger.info(
            f'You should see a "hello-world.html" opened in your browser, '
            f'if not you may open {url_html_path} manually'
        )

    colored_url = colored('https://opensource.jina.ai', color='cyan', attrs='underline')
    default_logger.info(
        f'🤩 Intrigued? Play with "jina hello fashion --help" and learn more about Jina at {colored_url}'
    )


def download_data(targets, download_proxy=None, task_name='download fashion-mnist'):
    """
    Download data.

    :param targets: target path for data.
    :param download_proxy: download proxy (e.g. 'http', 'https')
    :param task_name: name of the task
    """
    opener = urllib.request.build_opener()
    opener.addheaders = [('User-agent', 'Mozilla/5.0')]
    if download_proxy:
        proxy = urllib.request.ProxyHandler(
            {'http': download_proxy, 'https': download_proxy}
        )
        opener.add_handler(proxy)
    urllib.request.install_opener(opener)
    with ProgressBar(task_name=task_name) as t:
        for k, v in targets.items():
            if not os.path.exists(v['filename']):
                urllib.request.urlretrieve(
                    v['url'], v['filename'], reporthook=lambda *x: t.update_tick(0.01)
                )
            if k == 'index-labels' or k == 'query-labels':
                v['data'] = load_labels(v['filename'])
            if k == 'index' or k == 'query':
                v['data'] = load_mnist(v['filename'])


def load_mnist(path):
    """
    Load MNIST data

    :param path: path of data
    :return: MNIST data in np.array
    """

    with gzip.open(path, 'rb') as fp:
        return np.frombuffer(fp.read(), dtype=np.uint8, offset=16).reshape([-1, 28, 28])


def load_labels(path: str):
    """
    Load labels from path

    :param path: path of labels
    :return: labels in np.array
    """
    with gzip.open(path, 'rb') as fp:
        return np.frombuffer(fp.read(), dtype=np.uint8, offset=8).reshape([-1, 1])
